//
// Copyright (c) 2005-2011 StreamBase Systems, Inc. All rights reserved.
//
package com.streambase.sb.contrib.adapter.sbmonitor;

import java.lang.reflect.Method;
import java.util.Iterator;

import org.slf4j.Logger;

import com.streambase.sb.StreamBaseException;
import com.streambase.sb.StreamBaseRuntimeException;
import com.streambase.sb.Tuple;
import com.streambase.sb.TupleException;
import com.streambase.sb.adapter.InputAdapter;
import com.streambase.sb.client.StreamBaseURI;
import com.streambase.sb.client.URIException;
import com.streambase.sb.contrib.adapter.sbmonitor.StatusSchemaHelper.StatusAction;
import com.streambase.sb.contrib.adapter.sbmonitor.StatusSchemaHelper.StatusType;
import com.streambase.sb.monitor.ModuleInfo;
import com.streambase.sb.monitor.MonitorListener;
import com.streambase.sb.monitor.OperatorInfo;
import com.streambase.sb.monitor.Snapshot;
import com.streambase.sb.monitor.StreamBaseMonitor;
import com.streambase.sb.monitor.StreamInfo;
import com.streambase.sb.monitor.ThreadInfo;
import com.streambase.sb.operator.Operator;
import com.streambase.sb.operator.Parameterizable;
import com.streambase.sb.operator.PortCounts;
import com.streambase.sb.operator.TypecheckException;
import com.streambase.sb.util.Msg;
import com.streambase.sb.util.Util;

/**
 * This adapter uses the StreamBase Monitor API to get performance information from a StreamBase Server.
 */
public class SBMonitorAdapter extends InputAdapter implements Parameterizable, Runnable {

	public static final long serialVersionUID = 1290542829976L;
	private Logger logger;
	// Properties
	private String URI;
	private boolean remoteConnect;
	private int reconnectInterval;
	private boolean enableOperatorInfo;
	private boolean enableModuleInfo;
	private boolean enableThreadInfo;
	private boolean enableStreamInfo;
	private boolean enableSystemInfo;

	// Port numbers
	private int operatorInfoPort=0;
	private int moduleInfoPort=0;
	private int threadInfoPort=0;
	private int systemInfoPort=0;
	private int streamInfoPort=0;

	// getURI method was introduced in 7.0
	private Method getRuntimeEnvironmentMethod=null;
	private Method getURIMethod=null;
	
	private String displayName = "StreamBase Monitor Adapter";
	
	private StreamBaseMonitor sbmonitorClient=null;

	/**
	 * Constructor
	 */
	public SBMonitorAdapter() {
		super();
		logger = getLogger();
		setPortHints(0, 5);
		setDisplayName(displayName);
		setShortDisplayName(this.getClass().getSimpleName());
		setURI("");
		setRemoteConnect(false);
		setReconnectInterval(1);
	}

	/**
	 * Typecheck the adapter.
	 */
	public void typecheck() throws TypecheckException {

		// Make sure the URI is empty, or good
		if (!Util.isEmpty(getURI())) {
			try {
				new StreamBaseURI(getURI());
			} catch (URIException e) {
				logger.debug(getName(), e);
				throw new PropertyTypecheckException(SBMonitorAdapterBeanInfo.URI_PROPTERY_NAME, e);
			}
		} else if (remoteConnect) {
			throw new PropertyTypecheckException(SBMonitorAdapterBeanInfo.URI_PROPTERY_NAME,
					Msg.format("{0}: You must supply a URI for remote connections", getName()));
		}

		int portCount=0;
		setOutputSchema(portCount++, StatusSchemaHelper.SCHEMA);
		
		if (getEnableOperator()) {
			setOutputSchema(portCount, OperatorInfoSchemaHelper.SCHEMA);
			operatorInfoPort=portCount++;
		}
		
		if (getEnableModule()) {
			setOutputSchema(portCount, ModuleInfoSchemaHelper.SCHEMA);
			moduleInfoPort=portCount++;
		}
		
		if (getEnableThread()) {
			setOutputSchema(portCount, ThreadInfoSchemaHelper.SCHEMA);
			threadInfoPort=portCount++;
		}
		
		if (getEnableStream()) {
			setOutputSchema(portCount, StreamInfoSchemaHelper.SCHEMA);
			streamInfoPort=portCount++;
		}
		
		if (getEnableSystem()) {
			setOutputSchema(portCount, SystemInfoSchemaHelper.SCHEMA);
			systemInfoPort=portCount++;
		}
	}

	// Update the port count for studio
	public PortCounts getPortCounts() throws TypecheckException {
		int portCount=1;
		
		if (getEnableOperator())
			portCount++;
		if (getEnableModule())
			portCount++;
		if (getEnableThread())
			portCount++;
		if (getEnableStream())
			portCount++;
		if (getEnableSystem())
			portCount++;
		
		return new PortCounts(0, portCount);
    }
	
	// Helper method to send a status tuple
    public void sendStatusTuple(StatusType type, String object, StatusAction action, String message) {
    	Tuple t;
        try {
            t=StatusSchemaHelper.populateTuple(getOutputSchema(0).createTuple(), type, object, action, message);
        } catch (TupleException e) {
            logger.debug(getName(), e);
            logger.warn(getName() + ": Exception while creating status tuple", e);
            return;
        }
        try {
            sendOutput(0, t);
        } catch (StreamBaseException e) {
            logger.debug(getName(), e);
            logger.warn(getName() + ": Exception while sending status tuple", e);
        }
    }
    
	/**
	 * Initialize the adapter.
	 */
	public void init() throws StreamBaseException {
		super.init();

		// Register the object so it will be run as a thread managed by StreamBase.
		registerRunnable(this, true);

		// see if we should use the getURI method, or the system property
		try {
			getRuntimeEnvironmentMethod = Operator.class.getMethod("getRuntimeEnvironment");
			Class<?> [] clazzArray = Operator.class.getClasses();
			for (Class<?> c : clazzArray) {
				if (c.getName().equals("com.streambase.sb.operator.Operator$RuntimeEnvironment")) {
					getURIMethod = c.getMethod("getURI");
					break;
				}
			}
		} catch (Exception e) {
			logger.debug(Msg.format("{0}: Using streambase.uri system property", getName()));
		}
		
		if (getURIMethod == null)
			getRuntimeEnvironmentMethod=null;
	}

	/**
	 * Shutdown adapter
	 */
	public void shutdown() {
		if (sbmonitorClient != null) {
			sbmonitorClient.terminate();
		}
	}

	/**
	 * Main thread of the adapter. Used to make connection to server.
	 */
	public void run() {
		String sbURI=null;
		sbURI=getURI();
		if (!getRemoteConnect()) {
			// Get the uri of this server.
			sbURI=getLocalURI();
		}
		
		if (shouldRun())
			sendStatusTuple(StatusType.CONNECTION, getName(), StatusAction.INITIALIZING, sbURI);
		
		// get the streambase monitor client connected
		while (shouldRun()) {
			if (sbmonitorClient == null) {
				sendStatusTuple(StatusType.CONNECTION, getName(), StatusAction.CONNECTING, sbURI);
				try {
					sbmonitorClient = new StreamBaseMonitor(sbURI);
				} catch (StreamBaseException e) {
					logger.debug(getName(), e);
					sendStatusTuple(StatusType.CONNECTION, getName(), StatusAction.FAILED, sbURI + ": " + e.getMessage());
					logger.warn(Msg.format("{0}: Could not connect to {1}", getName(), sbURI));
					sbmonitorClient=null;
				}
				
				if (sbmonitorClient != null)  {
					if (!sbmonitorClient.isClosed()) {
						// For large number of operators detail takes a long while, and we don't use the data.
						sbmonitorClient.enableOperatorDetail(false);
						
				        sbmonitorClient.addMonitorListener(new StreamBaseMonitorAdapter());
				        
				        sendStatusTuple(StatusType.CONNECTION, getName(), StatusAction.CONNECTED, sbURI);
				        try {
							sbmonitorClient.run();
						} catch (Exception e) {
							if (shouldRun()) {
								logger.debug(getName(), e);
								logger.warn(Msg.format("{0}: Error trying to run monitor client on {1}", getName(), sbURI));
							} else {
								break;
							}
						}
						// Don't try and send a tuple if the server is shutting down
						if (shouldRun()) {
							sendStatusTuple(StatusType.CONNECTION, getName(), StatusAction.DISCONNECTED, sbURI);
						}
					}
					// returning from run means the server went away
					Util.close(sbmonitorClient);
					sbmonitorClient=null;
				}
			}
			
	        try {
				Thread.sleep(reconnectInterval*1000);
			} catch (InterruptedException e) {}
		}
	}

    /**
     * Return the URI string for the local server (i.e., the server this code is running it).
     */
    private String getLocalURI() {
    	String result=null;
    	
    	if (getURIMethod == null) {
    		result = Util.getSystemProperty(StreamBaseURI.URI_PROPERTY_NAME);
    	} else {
    		try {
				Object runtimeEnvironment=getRuntimeEnvironmentMethod.invoke(this);
				result = ((StreamBaseURI)getURIMethod.invoke(runtimeEnvironment)).toString();
			} catch (Exception e) {
				logger.debug(getName(), e);
			} 
    	}
        if (Util.isEmpty(result)) {
            throw new StreamBaseRuntimeException(Msg.format("{0}: Internal error: this server's URI is unexpectedly empty", getName()));
        }
        
        return result;
    }
    
	public class StreamBaseMonitorAdapter implements MonitorListener   {
		
	    public void snapshotReceived(Snapshot monSnap) {
	          
	    	// Go through the snapshot emit tuples of interest.
	    	if (getEnableOperator()) {
	            for (Iterator<OperatorInfo> iter = monSnap.operatorInfos(); iter.hasNext(); ) {
	            	try {
						sendOutputAsync(operatorInfoPort, OperatorInfoSchemaHelper.createOperatorTuple(monSnap.getTimestamp(), iter.next()));
					} catch (TupleException e) {
						logger.debug(getName(), e);
						logger.error(Msg.format("{0}: Unexpected error sending operator info", getName()), e);
					} catch (StreamBaseException e) {
						logger.debug(getName(), e);
						logger.error(Msg.format("{0}: Unexpected error sending operator info", getName()), e);
					}
	            }
	    	}
	    	
	    	if (getEnableModule()) {
	            for (Iterator<ModuleInfo> iter = monSnap.moduleInfos(); iter.hasNext(); ) {
	            	try {
						sendOutputAsync(moduleInfoPort, ModuleInfoSchemaHelper.createModuleTuple(monSnap.getTimestamp(), iter.next()));
					} catch (TupleException e) {
						logger.debug(getName(), e);
						logger.error(Msg.format("{0}: Unexpected error sending module info", getName()), e);
					} catch (StreamBaseException e) {
						logger.debug(getName(), e);
						logger.error(Msg.format("{0}: Unexpected error sending module info", getName()), e);
					}
	            }
	    	}
	    	
	    	if (getEnableThread()) {
	            for (Iterator<ThreadInfo> iter = monSnap.threadInfos(); iter.hasNext(); ) {
	            	try {
						sendOutputAsync(threadInfoPort, ThreadInfoSchemaHelper.createThreadTuple(monSnap.getTimestamp(), iter.next()));
					} catch (TupleException e) {
						logger.debug(getName(), e);
						logger.error(Msg.format("{0}: Unexpected error sending thread info", getName()), e);
					} catch (StreamBaseException e) {
						logger.debug(getName(), e);
						logger.error(Msg.format("{0}: Unexpected error sending thread info", getName()), e);
					}
	            }
	    	}
	    	
	    	if (getEnableStream()) {
	            for (Iterator<StreamInfo> iter = monSnap.streamInfos(); iter.hasNext(); ) {
	            	try {
						sendOutputAsync(streamInfoPort, StreamInfoSchemaHelper.createStreamTuple(monSnap.getTimestamp(), iter.next()));
					} catch (TupleException e) {
						logger.debug(getName(), e);
						logger.error(Msg.format("{0}: Unexpected error sending stream info", getName()), e);
					} catch (StreamBaseException e) {
						logger.debug(getName(), e);
						logger.error(Msg.format("{0}: Unexpected error sending stream info", getName()), e);
					}
	            }
	    	}
	    	
	    	if (getEnableSystem()) {
	        	try {
					sendOutputAsync(systemInfoPort, SystemInfoSchemaHelper.createSystemTuple(monSnap.getTimestamp(), monSnap.getSystemInfo()));
				} catch (TupleException e) {
					logger.debug(getName(), e);
					logger.error(Msg.format("{0}: Unexpected error sending system info", getName()), e);
				} catch (StreamBaseException e) {
					logger.debug(getName(), e);
					logger.error(Msg.format("{0}: Unexpected error sending system info", getName()), e);
				}
	    	}
	    }
	}

	/***************************************************************************************
	 * The getter and setter methods provided by the Parameterizable object.               *
	 * StreamBase Studio uses them to determine the name and type of each property         *
	 * and obviously, to set and get the property values.                                  *
	 ***************************************************************************************/

	public void setURI(String URI) {
		this.URI = URI;
	}
	public String getURI() {
		return this.URI;
	}
	public boolean shouldEnableURI() {
		return remoteConnect;
	}
	
	public void setRemoteConnect(boolean b) {
		this.remoteConnect = b;
	}
	public boolean getRemoteConnect() {
		return this.remoteConnect;
	}
	
	public void setReconnectInterval(int interval) {
		this.reconnectInterval = interval;
	}
	public int getReconnectInterval() {
		return this.reconnectInterval;
	}
	
	public void setEnableOperator(boolean b) {
		this.enableOperatorInfo = b;
	}
	public boolean getEnableOperator() {
		return this.enableOperatorInfo;
	}

	public void setEnableModule(boolean b) {
		this.enableModuleInfo = b;
	}
	public boolean getEnableModule() {
		return this.enableModuleInfo;
	}
	
	public void setEnableThread(boolean b) {
		this.enableThreadInfo = b;
	}
	public boolean getEnableThread() {
		return this.enableThreadInfo;
	}
	
	public void setEnableStream(boolean b) {
		this.enableStreamInfo = b;
	}
	public boolean getEnableStream() {
		return this.enableStreamInfo;
	}
	
	public void setEnableSystem(boolean b) {
		this.enableSystemInfo = b;
	}
	public boolean getEnableSystem() {
		return this.enableSystemInfo;
	}
}
